#include "Agent.h"

#include <iostream>
#include <QRegExp>
#include <QtDebug>

#include "Sensor.h"

#include <time.h>

const int Agent::TIMEOUT = 3000; // 3 sec

Agent::Agent() {

    this->uniqueID = QUuid::createUuid();
    
	this->model    = "USARBot.P2AT";
    this->name     = "Jimmy";
	this->position = "0,0,0";
	this->rotation = "0,0,0";

    this->isConnected = false;
    this->keepRunning = false;
    
    this->logDevice = NULL;
    this->logStream = NULL;
    this->logWrite = false;

	this->renderingmode = Agent::RENDERINGMODE_NONE;

	this->server.setQueueWhenDisconnected(true);
}

Agent::~Agent() {
	if(logStream != NULL)
		delete logStream;
}

void Agent::configure(QString model, QString name, QString position, QString rotation) {
    if (this->isConnected)
        throw Exception("Cannot configure agent when it is already connected");

    this->model    = model;
    this->name     = name;
	this->position = position;
	this->rotation = rotation;

	this->send(
		QString("INIT {ClassName %1} {Name %2} {Location %3} {Rotation %4}\r\n")
			.arg(model)
			.arg(name)
			.arg(position)
			.arg(rotation)
		);
}

void Agent::mount(Sensor* sensor) {
	
    QString type = sensor->getType();
	QString name = sensor->getName();

	if (!this->sensors.contains(type)) {
		QMap< QString, Sensor* > map;
		map.insert(name, sensor);
		this->sensors.insert(type, map);
	} else {
		QMap< QString, Sensor* >& map = this->sensors[type];
        if (map.contains(name))
            throw Exception("Sensor name must be unique");
		map.insert(name, sensor);
	}

	qDebug() << QString("Mounted %1 Sensor: %2").arg(type).arg(name);

    sensor->setAgent(this);

}

void Agent::unmount(Sensor* sensor) {
	
    QString type = sensor->getType();
	QString name = sensor->getName();

    if (!this->sensors.contains(type))
        throw Exception(QString("No Sensor of type %1 found").arg(type));

	QMap< QString, Sensor* >& map = this->sensors[type];
    if (!map.contains(name))
        throw Exception(QString("Sensor %1 not found").arg(name));
    map.remove(name);

    qDebug() << QString("Unmounted %1 Sensor: %2").arg(type).arg(name);

    sensor->setAgent(0);

}

QUuid Agent::getUniqueID() {
    return this->uniqueID;
}

QString Agent::getUniqueName() {
    return this->name + "_" + this->uniqueID.toString();
}

QString Agent::getName() {
    return this->name;
}

void Agent::connect(const QString &host, const quint16 port) {

	// NOT ALLOWED WHEN READING LOGFILE
	if( logStream != NULL && !logWrite)
		return;

	server.connect(host,port);
}

void Agent::disconnect() {
	this->stop();
	
	// NOT ALLOWED WHEN READING LOGFILE
	if( logStream != NULL && !logWrite)
		return;
	
	this->server.disconnect();
}

void Agent::stop() {
	this->keepRunning = false;
	wait();
}



void Agent::run() {

	QRegExp start("^(\\S+)\\s*");
	QRegExp bits("\\s*(\\{([^\\} ]+) ([^\\} ]+)\\}|\\{([^\\} ]+) ([^\\} ]+) ([^\\} ]+) ([^\\}]+)\\})\\s*");

    this->keepRunning = true;
    
    QString msg;
    
	QMap< QString, QString > parts;
    while (this->keepRunning) {

        // send out any pending commands
        if (this->keepRunning && server.isConnected()) {
			QString cmd;
			while (!commands.isEmpty()) {
				cmd = commands.dequeue();
				server.send(cmd);
				//qDebug() << "Sent: " << cmd;
			}
		}


		parts.clear();
		
		if( logStream != NULL && !logWrite )
		{
			if( logStream->atEnd() )
			{
				keepRunning = false;
				msg = "";
			}
			else
			{
				msg = logStream->readLine(0);
				// skip "commented" lines
				if(msg.startsWith("#"))
				{
					while( msg.startsWith("#") && !logStream->atEnd() )
						msg = logStream->readLine(0);

					if( logStream->atEnd() )
						keepRunning = false;
				}
			}
		}
		else
		{
			// recieve incoming message
			msg = server.receive(TIMEOUT);
		}

		if (msg.length() == 0)
			continue;
		
		//qDebug() << "Received: " << msg;

		if( logStream != NULL && logWrite ) {
			(*logStream) << msg << endl;	
		}

		if( msg.indexOf("RFID") > 0)
			qDebug() << msg;

        // process message
		if (this->keepRunning) {

            parts.clear();
			if (start.indexIn(msg) > -1) {

                parts.insert("MsgType", start.cap(1));
				
				int pos = start.matchedLength();
				while ((pos = bits.indexIn(msg, pos)) != -1 ) {
					if (bits.cap(2).length() != 0 ) {
						parts.insertMulti(bits.cap(2), bits.cap(3));
					} else if( bits.cap(4).length() != 0 ) {
						if( bits.cap(4) == "Status" ) // for victim RFID
						{
							QString val = bits.cap(5)+" "+bits.cap(6)+" "+bits.cap(7);
							parts.insertMulti(bits.cap(4), val);
						} else {
							parts.insertMulti(bits.cap(5), bits.cap(7));
						}
					}
					pos += bits.matchedLength();
				}

                this->recieve(&parts);
				
			} else {
				qDebug() << "Not valid message!:" << msg;
				continue;
			}
		}
	}
}

void Agent::logTo(QIODevice* log) {
	logDevice = log;
	logStream = new QTextStream(log);
	logWrite = true;
}


void Agent::logFrom(QIODevice* log) {
	logDevice = log;
	logStream = new QTextStream(log);
	logWrite = false;
}

void Agent::send(QString cmd) {
	QMutexLocker locker(&outputMutex);
	commands.enqueue(cmd);
}

void Agent::recieve(QMap<QString, QString>* parts) {

    if (!parts->contains("Type")) 
        return;

	QString type = parts->value("Type");
    QString name = Sensor::ANYNAME;
    if (parts->contains("Name")) 
        name = parts->value("Name");

    QMap< QString, Sensor* >& map = this->sensors[type];
    QMapIterator<QString, Sensor*> iterator(map);
    Sensor* sensor;
    QString sname;

    while (iterator.hasNext()) {
        iterator.next();
        sensor = iterator.value();
        sname  = sensor->getName();
        if (name == Sensor::ANYNAME || sname == Sensor::ANYNAME) {
            sensor->accept(parts);
        } else if (sname == name) {
            sensor->accept(parts);
        }
    }
}

void Agent::setRenderingMode(uint mode) {
	this->renderingmode = mode;
}
