#include "HoughTransformer.h"

#include <QtDebug>
#include <iostream>

const float HoughTransformer::SQRT_2_INV = 0.707106781186547;

HoughTransformer::HoughTransformer(float totalwidth, float totalheight, float resolution)
{
	width = 2 * (int)ceil( totalwidth /2 /resolution );
	height = (int)ceil( totalheight /resolution );
	
	hits = new int*[width];
	for(int i=0; i<width; i++)
	{
		hits[i] = new int[height];
	}
	
	qDebug() << "HoughTransform uses grid: " << width << height;
	
	res = resolution;
	
	threshold = 28;
	
	setRadius(.4); // estimated for victims
	
	SIN = COS = NULL;
	cacheLength = 0;
}

HoughTransformer::~HoughTransformer()
{
	for(int i=0; i<width; i++)
		delete[] hits[i];
	delete[] hits;
	
	if(COS != NULL)
		delete[] COS;
	if(SIN != NULL)
		delete[] SIN;
}

void HoughTransformer::reset()
{
	for(int i=0; i<width; i++)
	for(int j=0; j<height; j++)
	{
		hits[i][j] = 0;
	}
}

void HoughTransformer::setRadius(float rad)
{
	radius = (int)floor(rad / res);
}

Point2 HoughTransformer::getMidPoint(int y, int x)
{
	Point2 ret;
	
	ret.x = (1+x)*res - res/2;
	ret.y = (y-width/2)*res - res/2;
	
	//qDebug() << "converted " <<x<<y << " to " << ret.x << ret.y;
	
	return ret;
}

void HoughTransformer::cacheSines(float left, float step, int newLength)
{
	if(cacheLength == newLength)
		return;
		
	if( SIN != NULL )
	{
		delete[] SIN;
		delete[] COS;
	}
	
	SIN = new float[newLength];
	COS = new float[newLength];
	
	float a = left;
	for(int i=0; i<newLength; i++)
	{
		SIN[i] = sin(a);
		COS[i] = cos(a);
		a+=step;
	}
}

bool HoughTransformer::qualifies(int y, int x)
{
	Point2 p = getMidPoint(y,x);
	float a = atan2(p.y, p.x);
	
	if( fabs(a) > .4 ) // don't, we should see it with the rfid sensor
		return false;
		
	if( sqrt( p.x*p.x + p.y*p.y ) > 5 )
		return false; 
		
	return true;
}

QList< Point2 > HoughTransformer::getCircles(const LaserData& data)
{
	reset();
	
	QList<Point2> ret;
	float left = -data.fov/2;
	float step = data.resolution;
	
	cacheSines(left, step, data.length);
	
	//drawCircle( .7, -.5);
	//*
	for(int i=0; i<data.length; i++)
	{
		drawCircle( data.data[i] * COS[i], data.data[i] * SIN[i] );
	}
	//*/
	
	/*
	for(int i=0; i<width; i++)
	{
		for(int j=0; j<height; j++)
		{
			std::cerr << hits[i][j] << " ";
		}
		std::cerr << std::endl;
	}
	//*/
	
	
	Point2 p;
	float max = 0;
	for(int i=0; i<width; i++)
	{
		for(int j=0; j<height; j++)
		{
			if( qualifies(i,j) && hits[i][j] > max ) {
				max = hits[i][j];
				p = getMidPoint(i,j);
			}
			/*if( hits[i][j] > threshold )
			{
				ret.append( getMidPoint(i,j) );
			}*/
		}
	}
	
	qDebug() << "Max hits: " << max << " at " << p.x << p.y;
	
	// actually, thresholding is not so great, so just use the maximum
	//ret.clear();
	
	if( max >= threshold )
		ret.append(p);
	
	return ret;
}

void HoughTransformer::hit(int x, int y)
{
	if( x >= 0 && x < height && y >= 0 && y < width )
	{
		hits[y][x]++;
	}
}

void HoughTransformer::hit(int cx, int cy, int x, int y)
{
	hit(cx+x,cy+y);
	hit(cx+x,cy-y);
	hit(cx-x,cy+y);
	hit(cx-x,cy-y);
	
	hit(cx+y,cy+x);
	hit(cx+y,cy-x);
	hit(cx-y,cy+x);
	hit(cx-y,cy-x);
}

void HoughTransformer::drawCircle(float cx, float cy)
{
	int y = radius;
	int d = -radius;
	int x2m1 = -1;
	
	int thresh = radius * SQRT_2_INV;
	
	int cy_int = ceil(cy / res);
	cy_int = cy_int + width/2 - 1;
		
	int cx_int = ceil(cx / res);
	
	//qDebug() << "converted from " << cx<<cy << " to " << cx_int << cy_int;
	
	/*
	hit(cx_int,cy_int);
	return;
	//*/
	
	//qDebug() << "trans: " << cx << cy << " -> " << cx_int << cy_int;
	
	//hit(cx_int, cy_int, 0, radius);
	hit(cx_int, cy_int + radius);
	hit(cx_int, cy_int - radius);
	hit(cx_int + radius, cy_int);
	hit(cx_int - radius, cy_int);
	
	for(int x = 1; x <= thresh; x++)
	{
		x2m1 += 2;
		
		d+= x2m1;
		
		if (d>=0) 
		{
			y--;
			d -= (y<<1);  /* Must do this AFTER y-- */
		}
		hit(cx_int, cy_int, x, y);
	}
}
