#include "Manifold.h"

#include <QQueue>

const float Manifold::SUBMAP_MAXTRANSLATIONVARIANCE = 4.0F;
const float Manifold::SUBMAP_MAXROTATIONVARIANCE    = M_PI / 8.0F;


Manifold::Manifold() 
: Graph()
{
}
Manifold::~Manifold() {
}



bool Manifold::hasRFID(const QString& ID) {
    return this->rfidPatches.contains(ID);
}
bool Manifold::hasVictim(const QString& ID) {
    return this->victimPatches.contains(ID);
}

void Manifold::insertStartNode(QUuid id)
{
	startNodes.insert(id);
}

void Manifold::updateVictimStatus(QString ID, QString status)
{
	if(status == "None") // dont put in the empty status
		return;
		
	if(victimStatus.contains(ID))
	{
		QString curstatus = victimStatus[ID];
		if(curstatus.indexOf(status) < 0)
		{
			if(curstatus.length() > 0)
				curstatus+=", ";
			victimStatus[ID] = curstatus+status;
		}
	}
	else
	{
		victimStatus.insert(ID, status);
	}
}

void Manifold::insertPatch(ScanObservation* patch) {
    this->insertNode(new ManifoldNode(this, patch));
    
    if(this->nodeCount() == 1)
	    	startNodes.insert(patch->getUniqueID());

	QMapIterator<QString, RFIDData> rfids = patch->rfidIterator();
	while(rfids.hasNext())
	{
		rfids.next();
		rfidPatches[rfids.value().ID].append(patch->getUniqueID());
	}

	QMapIterator<QString, VictimRFIDData> victims = patch->victimIterator();
	while(victims.hasNext())
	{
		victims.next();
		victimPatches[victims.value().ID].append(patch->getUniqueID());
	}
}
void Manifold::insertRelation(ScanObservation* fromPatch, ScanObservation* toPatch, Matrix3 mcov) {
    ManifoldNode* fromNode = (ManifoldNode*) this->getNode(fromPatch->getUniqueID());
    ManifoldNode* toNode   = (ManifoldNode*) this->getNode(toPatch->getUniqueID());
    this->insertLink(new ManifoldLink(this, fromNode, toNode, mcov));
}

void Manifold::insertRelation(const QUuid& fromPatch, const QUuid& toPatch, Matrix3 mcov) {
    ManifoldNode* fromNode = (ManifoldNode*) this->getNode(fromPatch);
    ManifoldNode* toNode   = (ManifoldNode*) this->getNode(toPatch);
    this->insertLink(new ManifoldLink(this, fromNode, toNode, mcov));
}

// sorry for the unorthodox placing
struct ManifoldNodeWithCovariance
{
	ManifoldNode* node;
	Matrix3 cov;
	
	ManifoldNodeWithCovariance(ManifoldNode* n)
	{
		node=n;
		cov(0,0) = 0; // set zero matrix
		cov(1,1) = 0;
		cov(2,2) = 0;
	}
	
	QUuid getId() {
		return node->patch->getUniqueID();
	}
};

bool operator<(const ManifoldNodeWithCovariance& left, const ManifoldNodeWithCovariance& right)
{
	float leftSum=0, rightSum=0;
	for(int i=0; i<3; i++)
		for(int j=0; j<3; j++)
		{
			leftSum += left.cov(i,j);
			rightSum += right.cov(i,j);
		}
		
	return leftSum < rightSum;
}

struct UncertainVector2
{
	Vector2 pose;
	Matrix3 cov;
	
	UncertainVector2( Vector2 p, Matrix3 c )
	{
		pose = p;
		cov = c;
	}
	
	float sumCov()
	{
		float sum=0;
		for(int i=0; i<3; i++)
			for(int j=0; j<3; j++)
			{
				sum += cov(i,j);
			}
			if (sum == 0) {
				return 1.0f/10000.0f;
			} else {
				return 1.0f/sum;
			}
	}
};


QList< RFIDData > Manifold::getRFIDs()
{
	QQueue< ManifoldNodeWithCovariance > nodeQueue;
	QSet< QUuid > alreadyVisited;
	
	QSetIterator< QUuid > startNodeIt(startNodes);
	while( startNodeIt.hasNext() ) {
		ManifoldNodeWithCovariance node((ManifoldNode*)this->getNode(startNodeIt.next()) );
		nodeQueue.enqueue( node );
	}
	
	QMap< QString, QList< UncertainVector2 > > rfidPoses;
	
	ManifoldNodeWithCovariance node = NULL;
	while( !nodeQueue.isEmpty() ) {
		node = nodeQueue.dequeue();
		
		//qDebug() << "visiting " << node.node->getUniqueID() << ", already visited? " << alreadyVisited.contains( node.node->getUniqueID() );
		
		Matrix3 toGlobal;
		//toGlobal.fromPose2D( node.node->patch->pose );
		toGlobal.rotate(node.node->patch->pose.rot);
		toGlobal.translate(node.node->patch->pose.x/1000, node.node->patch->pose.y/1000);
		
		QMapIterator< QString, RFIDData > patchrfids = node.node->patch->rfidIterator();
		while( patchrfids.hasNext() )
		{
			patchrfids.next();
			RFIDData rfiddata = patchrfids.value();
			
			Vector2 pos( rfiddata.dx, rfiddata.dy );
			pos = toGlobal.multiply(pos);
			
			//qDebug() << "seen rfid at " << pos.x << pos.y << " on patch at " << node.node->patch->pose.x/1000 << node.node->patch->pose.y/1000 << node.node->patch->pose.rot;
			
			rfidPoses[rfiddata.ID].append( UncertainVector2( pos, node.cov ) ); 
		}
		
		alreadyVisited.insert( node.node->getUniqueID() );
		
		// enqueue next links
		QMapIterator<QUuid, GraphLink*> links = node.node->linkIterator();
		
		while(links.hasNext()) {
			links.next();
			if( !alreadyVisited.contains( links.value()->getOtherNode( node.node )->getUniqueID() ) )
			{
				ManifoldNodeWithCovariance nextNode((ManifoldNode*)links.value()->getOtherNode( node.node ));
				nextNode.cov = node.cov + ((ManifoldLink*)links.value())->mcov;
				nodeQueue.enqueue(nextNode);
			}
		}
		
		qSort( nodeQueue );
	}
	
	
	QList< RFIDData > rfids;
	// now do weighted average
	QList<QString> keys = rfidPoses.keys();
	for( int i=0; i<keys.size(); i++)
	{
		QString ID = keys[i];
		
		RFIDData data;
		data.ID = ID;
		data.dx = data.dy = data.dz = 0;
		
		QList< UncertainVector2 > possiblePos = rfidPoses.value(ID);
		float norm=0;
		for(int k=0; k<possiblePos.size(); k++)
		{
			norm += possiblePos[k].sumCov();
		}
		
		for(int k=0; k<possiblePos.size(); k++)
		{
			data.dx += possiblePos[k].pose.x * ( possiblePos[k].sumCov() / norm);
			data.dy += possiblePos[k].pose.y * ( possiblePos[k].sumCov() / norm);
			
			//qDebug() << "rfid :" << data.ID << " ? at pos:" << possiblePos[k].pose.x << possiblePos[k].pose.y;
		}
		/*
		if(possiblePos.size() == 1)
		{
			data.dx = possiblePos[0].pose.x;
			data.dy = possiblePos[0].pose.y;
		}*/
		
		//qDebug() << "average: " << data.ID << " -> " << data.dx << data.dy;
		
		rfids.append(data);
	}
	
	return rfids;
}

QList< VictimRFIDData > Manifold::getVictimRFIDs()
{
	QQueue< ManifoldNodeWithCovariance > nodeQueue;
	QSet< QUuid > alreadyVisited;
	
	QSetIterator< QUuid > startNodeIt(startNodes);
	while( startNodeIt.hasNext() ) {
		ManifoldNodeWithCovariance node((ManifoldNode*)this->getNode(startNodeIt.next()) );
		nodeQueue.enqueue( node );
	}
	
	QMap< QString, QList< UncertainVector2 > > rfidPoses;
	
	ManifoldNodeWithCovariance node = NULL;
	while( !nodeQueue.isEmpty() ) {
		node = nodeQueue.dequeue();
		
		//qDebug() << "visiting " << node.node->getUniqueID() << ", already visited? " << alreadyVisited.contains( node.node->getUniqueID() );
		
		Matrix3 toGlobal;
		//toGlobal.fromPose2D( node.node->patch->pose );
		toGlobal.rotate(node.node->patch->pose.rot);
		toGlobal.translate(node.node->patch->pose.x/1000, node.node->patch->pose.y/1000);
		
		QMapIterator< QString, VictimRFIDData > patchrfids = node.node->patch->victimIterator();
		while( patchrfids.hasNext() )
		{
			patchrfids.next();
			VictimRFIDData rfiddata = patchrfids.value();
			
			Vector2 pos( rfiddata.dx, rfiddata.dy );
			pos = toGlobal.multiply(pos);
			
			rfidPoses[rfiddata.ID].append( UncertainVector2( pos, node.cov ) ); 
		}
		
		alreadyVisited.insert( node.node->getUniqueID() );
		
		// enqueue next links
		QMapIterator<QUuid, GraphLink*> links = node.node->linkIterator();
		
		while(links.hasNext()) {
			links.next();
			if( !alreadyVisited.contains( links.value()->getOtherNode( node.node )->getUniqueID() ) )
			{
				ManifoldNodeWithCovariance nextNode((ManifoldNode*)links.value()->getOtherNode( node.node ));
				nextNode.cov = node.cov + ((ManifoldLink*)links.value())->mcov;
				nodeQueue.enqueue(nextNode);
			}
		}
		
		qSort( nodeQueue );
	}
	
	
	// now do weighted average
	QList< VictimRFIDData > rfids;
	QList<QString> keys = rfidPoses.keys();
	for( int i=0; i<keys.size(); i++)
	{
		QString ID = keys[i];
		
		VictimRFIDData data;
		data.ID = ID;
		data.dx = data.dy = data.dz = 0;
		
		QList< UncertainVector2 > possiblePos = rfidPoses.value(ID);
		float norm=0;
		for(int k=0; k<possiblePos.size(); k++)
		{
			norm += possiblePos[k].sumCov();
		}
		
		for(int k=0; k<possiblePos.size(); k++)
		{
			data.dx += possiblePos[k].pose.x / possiblePos.size();
			data.dy += possiblePos[k].pose.y / possiblePos.size();
		}
		/*if(possiblePos.size() == 1)
		{
			data.dx = possiblePos[0].pose.x;
			data.dy = possiblePos[0].pose.y;
		}*/
		
		data.status = getVictimStatus(data.ID);
		
		rfids.append(data);
	}


	// now check for additional victims on disconnected parts
	QMapIterator< QString, QList<QUuid> > victimPatchesIt(this->victimPatches);
	while (victimPatchesIt.hasNext()) {
		victimPatchesIt.next();
		QString ID = victimPatchesIt.key();
		if (keys.contains(ID)) {
			// this victim was already found in the breadth first run, skip
			continue;
		}

		// otherwise: report it as well ...
		VictimRFIDData data;
		data.ID = ID;
		data.dx = data.dy = data.dz = 0;

		QList<QUuid> nodes = victimPatchesIt.value();
		for (int i = 0; i < nodes.count(); i++) {
			ManifoldNode*  node = (ManifoldNode*) this->getNode(nodes[i]);
			VictimRFIDData rfid = node->patch->getVictim(ID);

			Matrix3 toGlobal;
			toGlobal.fromPose2D(node->patch->pose);

			Vector2 vpose(rfid.dx, rfid.dy);
			vpose = toGlobal.multiply(vpose);

			data.dx += vpose.x / nodes.size();
			data.dy += vpose.y / nodes.size();
		}
		
		data.dx/=1000.0f;
		data.dy/=1000.0f;

		rfids.append(data);
	}
	
	return rfids;
}

QList<QUuid> Manifold::getRFIDPatches(QString ID) {
	return this->rfidPatches[ID];
}
QList<QUuid> Manifold::getVictimPatches(QString ID) {
	return this->victimPatches[ID];
}

Manifold* Manifold::getSubMap(ScanObservation* patch) {

    Manifold* submap = new Manifold();
    QList<QUuid> done;
    Matrix3 sumcov(0.0F);

    ManifoldNode* node = (ManifoldNode*) this->getNode(patch->getUniqueID());
    done.append(node->getUniqueID());
    node->copyTo(submap);

    this->getSubMap(submap, done, sumcov, node);

    return submap;
}
    
void Manifold::getSubMap(Manifold* submap, QList<QUuid>& done, Matrix3 sumcov, ManifoldNode* node) {

    QMapIterator<QUuid, GraphLink*> iterator = node->linkIterator();
    QList<ManifoldLink*> links;
    ManifoldLink* nextLink;
    ManifoldNode* nextNode;

    Matrix3 cov;
    bool ok;

    while (iterator.hasNext()) {

        iterator.next();
        nextLink = (ManifoldLink*) iterator.value();
        nextNode = (ManifoldNode*) nextLink->getOtherNode(node);

        if (!done.contains(nextNode->getUniqueID())) {

            done.append(nextNode->getUniqueID());

            cov = sumcov + nextLink->mcov;

            // check covariance
            ok = true;
            ok = ok && (1.96 * sqrt(cov(0, 0)) < Manifold::SUBMAP_MAXTRANSLATIONVARIANCE);
            ok = ok && (1.96 * sqrt(cov(1, 1)) < Manifold::SUBMAP_MAXTRANSLATIONVARIANCE);
            ok = ok && (1.96 * sqrt(cov(2, 2)) < Manifold::SUBMAP_MAXROTATIONVARIANCE);

            if (ok) {
                nextNode->copyTo(submap);
                nextLink->copyTo(submap);
                links.append(nextLink);
            }
        }
    }

    for (int i = 0; i < links.count(); i++) {
        nextLink = (ManifoldLink*) links.at(i);
        nextNode = (ManifoldNode*) nextLink->getOtherNode(node);
        this->getSubMap(submap, done, sumcov + nextLink->mcov, nextNode);
    }
}


Manifold* Manifold::getConnectedSubSet() {

    Manifold* subset = new Manifold();
    QList<QUuid> done;

	QSetIterator< QUuid > startNodeIterator(this->startNodes);
    ManifoldNode* node;
	while( startNodeIterator.hasNext() ) {
		QUuid nodeId = startNodeIterator.next();
		subset->insertStartNode( nodeId );
		node = (ManifoldNode*) this->getNode(nodeId);
		done.append(node->getUniqueID());
		node->copyTo(subset);
		this->getConnectedSubSet(subset, done, node);
	}

    return subset;
}

Manifold* Manifold::getConnectedSubSet(ManifoldNode* node) {
    Manifold* subset = new Manifold();
    QList<QUuid> done;

	done.append(node->getUniqueID());
	node->copyTo(subset);

	this->getConnectedSubSet(subset, done, node);

    return subset;

}

void Manifold::getConnectedSubSet(Manifold* subset, QList<QUuid>& done, ManifoldNode* node) {

	QMapIterator<QUuid, GraphLink*> iterator = node->linkIterator();
    QList<ManifoldLink*> links;
    ManifoldLink* nextLink;
    ManifoldNode* nextNode;

    while (iterator.hasNext()) {
        iterator.next();
        nextLink = (ManifoldLink*) iterator.value();
        nextNode = (ManifoldNode*) nextLink->getOtherNode(node);

        if (!done.contains(nextNode->getUniqueID())) {
            done.append(nextNode->getUniqueID());
            nextNode->copyTo(subset);
            nextLink->copyTo(subset);
            links.append(nextLink);
        }
    }

    for (int i = 0; i < links.count(); i++) {
        nextLink = (ManifoldLink*) links.at(i);
        nextNode = (ManifoldNode*) nextLink->getOtherNode(node);
		this->getConnectedSubSet(subset, done, nextNode);
    }
}

Bounds Manifold::computeBounds() {

    QMapIterator<QUuid, GraphNode*> iterator = this->nodeIterator();
    ManifoldNode* node;
	ScanObservation* patch;

    Bounds bounds;
    bounds.top = bounds.left = bounds.height = bounds.width = 0;

    Bounds pbounds;

    while (iterator.hasNext()) {
        iterator.next();
        node = (ManifoldNode*) iterator.value();
		patch = node->patch;
        pbounds.left   = patch->pose.x - patch->maxdistance;
        pbounds.top    = patch->pose.y - patch->maxdistance;
		pbounds.height = 2 * patch->maxdistance;
		pbounds.width  = 2 * patch->maxdistance;
		
		//qDebug() << "pose: " << patch->pose.x << patch->pose.y << patch->pose.rot;
		
        bounds = bounds + pbounds;
    }

    return bounds;
}
void Manifold::render(ManifoldRenderer* renderer) {

    QMapIterator<QUuid, GraphNode*> nodeIterator = this->nodeIterator();
    ManifoldNode* node1;
	uint i = 1;
    while (nodeIterator.hasNext()) {
        nodeIterator.next();
        node1 = (ManifoldNode*) nodeIterator.value();
		renderer->renderScan(node1->patch->pose, node1->patch, i);
		i++;
    }
    
    QMapIterator<QUuid, GraphLink*> linkIterator = this->linkIterator();
    ManifoldLink* link;
    ManifoldNode* node2;
    while( linkIterator.hasNext()) {
    		linkIterator.next();
    		link  = (ManifoldLink*) linkIterator.value();
    		node1 = (ManifoldNode*) link->node1();
    		node2 = (ManifoldNode*) link->node2();
		renderer->renderLink(node1->patch->pose, node2->patch->pose);
    }
    
    nodeIterator.toFront();
    while (nodeIterator.hasNext()) {
        nodeIterator.next();
        node1 = (ManifoldNode*) nodeIterator.value();
		renderer->renderScansRFIDs(node1->patch->pose, node1->patch, i);
		i++;
    }
    
    QList<RFIDData> rfids = this->getRFIDs();
	RFIDData rfid;
    for(int i=0; i<rfids.size(); i++)
    {
		rfid = rfids[i];
    	renderer->renderRFID(rfid);
    }
    
    QList<VictimRFIDData> vicrfids = this->getVictimRFIDs();
    for(int i=0; i<vicrfids.size(); i++)
    {
    		renderer->renderVictimRFID(vicrfids[i]);
    }
}


#ifdef USE_OPENGL
void Manifold::drawGL() {
	//QMapIterator<QString, Trail*> it(trails);
	//
	//glColor3f(1,1,1);
	//
	//glPushMatrix();
	//
	//while(it.hasNext()) {
	//	it.next();
	//	it.value()->drawGL();
	//}
	//
	//glColor3f(1,1,1);
	//octree.drawGL();
	//
	//glColor3f(1,.2,.2);
	//
	//glBegin(GL_LINES);
	//	glVertex3d(1,1,1);
	//	glVertex3d(2,2,4);
	//	
	//	glVertex3d(2,2,4);
	//	glVertex3d(2,1,0);
	//	
	//	glVertex3d(2,1,0);
	//	glVertex3d(1,1,1);
	//glEnd();
	//
	//glPopMatrix();
}

void Manifold::setColors(QColor trail, QColor scans) {
	//Trail* trailObj = NULL;
	//if( trails.contains(trailName) ) {
	//	trailObj = trails[trailName];
	//	trailObj->setColors(trail,scans);
	//}
}
#endif

QDataStream& operator>>(QDataStream& out, Manifold&)
{
	return out;
}
QDataStream& operator<<(QDataStream& in, const Manifold&)
{
	return in;
}


ManifoldNode::ManifoldNode(Manifold* manifold, ScanObservation* patch) 
: GraphNode(manifold, patch->getUniqueID())
{
    this->patch = patch;
}
ManifoldNode::~ManifoldNode() {
}

void ManifoldNode::copyTo(Manifold* manifold) {
    manifold->insertPatch(this->patch);
}

void ManifoldNode::moveAllConnectedBy(Vector2 center, Pose2D dpose) {
	QQueue< ManifoldNode* > nodeQueue;
	QSet< QUuid > alreadyVisited;
	
	nodeQueue.enqueue(this);
	
	ManifoldNode* node = NULL;
	while( !nodeQueue.isEmpty() ) {
		node = nodeQueue.dequeue();
		
		//qDebug() << "visiting " << node.node->getUniqueID() << ", already visited? " << alreadyVisited.contains( node.node->getUniqueID() );
		
		node->moveBy(center,dpose);
		
		alreadyVisited.insert( node->getUniqueID() );
		
		// enqueue next links
		QList<GraphLink*> links = node->links.values();
		
		for(int i=0; i<links.size(); i++) {
			if( !alreadyVisited.contains( links[i]->getOtherNode( node )->getUniqueID() ) )
			{
				nodeQueue.enqueue((ManifoldNode*)links[i]->getOtherNode( node ));
			}
		}
	}
}

void ManifoldNode::moveBy(Vector2 center, Pose2D dpose) {

	Pose2D ownPose = this->patch->pose;
	Vector2 diff(ownPose.x-center.x, ownPose.y-center.y );

	ownPose.x = (center.x+dpose.x);
	ownPose.y = (center.y+dpose.y);

	Matrix3 rotmx;
	rotmx.setRotation(dpose.rot);
	diff = rotmx.multiply(diff);

	ownPose.x += diff.x;
	ownPose.y += diff.y;

	ownPose.rot += dpose.rot;

	this->patch->pose = ownPose;
}




ManifoldLink::ManifoldLink(Manifold* manifold, ManifoldNode* fromNode, ManifoldNode* toNode, const Matrix3& mcov) 
: GraphLink(manifold, QUuid::createUuid(), fromNode, toNode)
{
    this->mcov = mcov;
}

ManifoldLink::~ManifoldLink() {
}
void ManifoldLink::copyTo(Manifold* manifold) {
    ManifoldNode* fromNode = (ManifoldNode*) manifold->getNode(this->fromNodeID);
    ManifoldNode* toNode   = (ManifoldNode*) manifold->getNode(this->toNodeID);
    manifold->insertLink(new ManifoldLink(manifold, fromNode, toNode, this->mcov));
}
