#include "Octree.h"

#include <QDebug>

// some standard headers
#include <math.h>
#include <limits>

// default minimum edge length, determines resolution of octree
float Octree::minEdgeLength = .02f; // in meters

// threshold to say if this cell is free
const float Octree::THRESHOLD_IS_FREE = .1f;
const float Octree::FLOAT_INF = std::numeric_limits<float>::infinity();

Octree::Octree(float edgeLength)
{
	this->edgeLength = minEdgeLength * pow( 2.0F,  floor( log( edgeLength / minEdgeLength ) / log(2.0F) ) + 1.0F);
	this->occupancy = .5f;
	this->hasVictim = .5f;
	this->hasRFID = .5f;
	parent = NULL;
	
	displayList = 0;
	isHit = false;
	
	for(int i=0; i<8; i++)
		children[i] = NULL;
}

Octree::Octree(float edgeLength, float minEdgeLength)
{
	Octree::minEdgeLength = minEdgeLength;
	
	this->edgeLength = minEdgeLength * pow( 2.0F, floor( log( edgeLength / minEdgeLength ) / log(2.0F) ) + 1.0F);
	this->occupancy = .5f;
	this->hasVictim = .5f;
	this->hasRFID = .5f;
	parent = NULL;
	
	displayList = 0;
	isHit = false;
	
	for(int i=0; i<8; i++)
		children[i] = NULL;
}

Octree::Octree(float edgeLength, Point3 center, Octree* parent)
{
	this->edgeLength = edgeLength;
	this->occupancy = .5f;
	this->hasVictim = .5f;
	this->hasRFID = .5f;
	this->parent = parent;
	
	this->center = center;
	
	displayList = 0;
	isHit = false;
	
	for(int i=0; i<8; i++)
		children[i] = NULL;
}

Octree::Octree(float edgeLength, float minEdgeLength, Point3 center)
{
	Octree(edgeLength, minEdgeLength);
	this->center = center;
}

Octree::~Octree()
{
	for(int i=0; i<8; i++)
		if( children[i] != NULL )
			delete children[i];
}

void Octree::setParent(Octree* parent)
{
	this->parent = parent;
}

bool Octree::isFree()
{
	return ( occupancy <= THRESHOLD_IS_FREE );
}

bool Octree::isLeaf()
{
	for(int i=0; i<8; i++)
		if( children[i] != NULL )
			return false;
	return true;
}

Octree* Octree::getChild(int i)
{
	if( i>0 && i<8 )
		return children[i];
	return NULL;
}

void Octree::insertVictim(Point3& p)
{
	// should we be a leaf?
	if( edgeLength <= minEdgeLength )
	{
		hasVictim = (hasVictim + 1)/2;
			
		return;	
	}
	
	// we still have more levels to make/iterate through
	
	int child = 0;
	if( p.x > center.x )
		child |= 4;
	if( p.y > center.y )
		child |= 2;
	if( p.z > center.z )
		child |= 1;
		
	if(children[child] == NULL)
	{
		insertChild(child);
	}
	
	children[child]->insertVictim(p);
}

void Octree::insertChild(char child)
{
	if(children[child] == NULL)
	{
		Point3 newCenter = center;
		newCenter.x -= edgeLength/4;
		newCenter.y -= edgeLength/4;
		newCenter.z -= edgeLength/4;
		
		if( (child & 4) != 0 )
			newCenter.x += edgeLength/2;
		if( (child & 2) != 0 )
			newCenter.y += edgeLength/2;
		if( (child & 1) != 0 )
			newCenter.z += edgeLength/2;
		
		children[child] = new Octree(edgeLength/2, newCenter, this);
	}
}

void Octree::drawScanLine(Ray ray)
{
	Vector3 t0, t1;
	
	ray.ordering = 0;
	if( ray.direction.x < 0.0 )
	{
		ray.origin.x = 2 * center.x - ray.origin.x;
		ray.direction.x = -ray.direction.x;
		ray.ordering |= 4;
	}
	if( ray.direction.y < 0.0 )
	{
		ray.origin.y = 2 * center.y - ray.origin.y;
		ray.direction.y = -ray.direction.y;
		ray.ordering |= 2;
	}
	if( ray.direction.z < 0.0 )
	{
		ray.origin.z = 2 * center.z - ray.origin.z;
		ray.direction.z = -ray.direction.z;
		ray.ordering |= 1;
	}
	
	float length_2 = edgeLength/2;
	
	t0 = Vector3(center.x-length_2, center.y-length_2, center.z-length_2);
	t0 = (t0 - ray.origin)/ray.direction;
	
	t1 = Vector3(center.x+length_2, center.y+length_2, center.z+length_2);
	t1 = (t1 - ray.origin)/ray.direction;
	
	if( t0.fmax() < t1.fmin() )
	{
		recursiveDrawScanLine(ray, t0, t1);
	}
}

void Octree::recursiveDrawScanLine(Ray& ray, Vector3& t0, Vector3& t1)
{
	// check for lines parallel to axis
	if( ray.direction.x == 0 && !( t0.x == -std::numeric_limits<float>::infinity() && t1.x == std::numeric_limits<float>::infinity() ) )
		return;
	if( ray.direction.y == 0 && !( t0.y == -std::numeric_limits<float>::infinity() && t1.y == std::numeric_limits<float>::infinity() ) )
		return;
	if( ray.direction.z == 0 && !( t0.z == -std::numeric_limits<float>::infinity() && t1.z == std::numeric_limits<float>::infinity() ) )
		return;

	// not hit by line
	if( t0.fmax() >= t1.fmin() )
		return;

	// not within line segment
	if( t1.fmin() < 0 || t0.fmax() > ray.length )
		return;
		
	//qDebug() << "Ray from "<< ray.origin.x << ray.origin.y << ray.origin.z << " hits cube at " << center.x << center.y << center.z;
	
	// ray comes through and we are already free and we don't intersect with the end bit of the ray
	//if( isLeaf() && isFree() && t1.fmin() < ray.length - ray.noise )
	//	return;
	
	if( edgeLength <= minEdgeLength )
	{
		isHit = true;
		return;
	}
	
	Vector3 tm = (t0 + t1)/2;

	// make sure that we are on the line for the dimension that we are parallel to
	if( ray.direction.x == 0 )
	{
		if( ray.origin.x < center.x )
			tm.x = FLOAT_INF;
		else
			tm.x = -FLOAT_INF;
	}
	if( ray.direction.y == 0 )
	{
		if( ray.origin.y < center.y )
			tm.y = FLOAT_INF;
		else
			tm.y = -FLOAT_INF;
	}
	if( ray.direction.z == 0 )
	{
		if( ray.origin.z < center.z )
			tm.z = FLOAT_INF;
		else
			tm.z = -FLOAT_INF;
	}

	char cur = firstNode(t0,tm);
	
	Vector3 left, right;
	while(cur < 8)
	{
		if( children[cur^ray.ordering] == NULL )
			insertChild(cur^ray.ordering);
			
		combine(cur, left, t0, tm);
		combine(cur, right, tm, t1);
		
		children[cur^ray.ordering]->recursiveDrawScanLine(ray, left, right);
		
		cur = nextNode(cur, right);
	}
}

Vector3& Octree::combine(char i, Vector3& in, Vector3& left, Vector3& right)
{
	in = left;
	if((i & 4) != 0)
		in.x = right.x;
	if((i & 2) != 0)
		in.y = right.y;
	if((i & 1) != 0)
		in.z = right.z;
	
	return in;
}

char Octree::firstNode(Vector3& t0, Vector3& tm)
{
	char n = 0;
	
	switch ( t0.minIndex() )
	{
		case 0:
			if( tm.y < t0.x )
				n |= 2;
			if( tm.z < t0.x )
				n |= 4;
			break;
		case 1:
			if( tm.x < t0.y )
				n |= 1;
			if( tm.z < t0.y )
				n |= 4;
			break;
		case 2:
			if( tm.x < t0.z )
				n |= 1;
			if( tm.y < t0.z )
				n |= 2;
	};
	
	return n;
}

char Octree::nextNode(char current, Vector3& exits)
{
	char mask = 0;
	switch ( exits.minIndex() )
	{
		case 0:
			mask |= 4;
			break;
		case 1:
			mask |= 2;
			break;
		case 2:
			mask |= 1;
	};
	
	// end condition
	if( (current & mask) != 0 ) {
		return 8;
	}
	
	return current | mask;
}

#ifdef USE_OPENGL
void Octree::drawGL()
{
	if(false)
	{
		float length_2 = edgeLength/2;
		glBegin(GL_LINES);
			glVertex3f( center.x - length_2, center.y - length_2, center.z - length_2 );
			glVertex3f( center.x + length_2, center.y - length_2, center.z - length_2 );
			
			glVertex3f( center.x - length_2, center.y + length_2, center.z - length_2 );
			glVertex3f( center.x + length_2, center.y + length_2, center.z - length_2 );
			
			glVertex3f( center.x - length_2, center.y - length_2, center.z + length_2 );
			glVertex3f( center.x + length_2, center.y - length_2, center.z + length_2 );
			
			glVertex3f( center.x - length_2, center.y + length_2, center.z + length_2 );
			glVertex3f( center.x + length_2, center.y + length_2, center.z + length_2 );
			
			
			glVertex3f( center.x - length_2, center.y - length_2, center.z - length_2 );
			glVertex3f( center.x - length_2, center.y + length_2, center.z - length_2 );
			
			glVertex3f( center.x - length_2, center.y - length_2, center.z + length_2 );
			glVertex3f( center.x - length_2, center.y + length_2, center.z + length_2 );
			
			glVertex3f( center.x + length_2, center.y - length_2, center.z - length_2 );
			glVertex3f( center.x + length_2, center.y + length_2, center.z - length_2 );
			
			glVertex3f( center.x + length_2, center.y - length_2, center.z + length_2 );
			glVertex3f( center.x + length_2, center.y + length_2, center.z + length_2 );
			
			
			glVertex3f( center.x - length_2, center.y - length_2, center.z - length_2 );
			glVertex3f( center.x - length_2, center.y - length_2, center.z + length_2 );
			
			glVertex3f( center.x - length_2, center.y + length_2, center.z - length_2 );
			glVertex3f( center.x - length_2, center.y + length_2, center.z + length_2 );
			
			glVertex3f( center.x + length_2, center.y - length_2, center.z - length_2 );
			glVertex3f( center.x + length_2, center.y - length_2, center.z + length_2 );
			
			glVertex3f( center.x + length_2, center.y + length_2, center.z - length_2 );
			glVertex3f( center.x + length_2, center.y + length_2, center.z + length_2 );
		glEnd();
	}
	
	if( isLeaf() )
	{
		if( isHit /*!isFree()*/ )
		{
			// lazy generation of display list
			if(!displayList)
			{
				displayList = glGenLists(1);
				glNewList(displayList, GL_COMPILE_AND_EXECUTE);
				
				if( true /*edgeLength < 4 * minEdgeLength*/ )
				{
					float length_10 = edgeLength*9/20;
					glBegin(GL_POINTS);
						glVertex3f( center.x - length_10, center.y - length_10, center.z - length_10 );
						glVertex3f( center.x - length_10, center.y - length_10, center.z + length_10 );
						glVertex3f( center.x - length_10, center.y + length_10, center.z - length_10 );
						glVertex3f( center.x - length_10, center.y + length_10, center.z + length_10 );
						glVertex3f( center.x + length_10, center.y - length_10, center.z - length_10 );
						glVertex3f( center.x + length_10, center.y - length_10, center.z + length_10 );
						glVertex3f( center.x + length_10, center.y + length_10, center.z - length_10 );
						glVertex3f( center.x + length_10, center.y + length_10, center.z + length_10 );
					glEnd();
				}
				else
				{
					float length_2 = edgeLength/2;
					glBegin(GL_QUADS);
						glVertex3f( center.x - length_2, center.y - length_2, center.z - length_2 );
						glVertex3f( center.x - length_2, center.y - length_2, center.z + length_2 );
						glVertex3f( center.x - length_2, center.y + length_2, center.z + length_2 );
						glVertex3f( center.x - length_2, center.y + length_2, center.z - length_2 );
						
						glVertex3f( center.x + length_2, center.y - length_2, center.z - length_2 );
						glVertex3f( center.x + length_2, center.y - length_2, center.z + length_2 );
						glVertex3f( center.x + length_2, center.y + length_2, center.z + length_2 );
						glVertex3f( center.x + length_2, center.y + length_2, center.z - length_2 );
						
						glVertex3f( center.x - length_2, center.y - length_2, center.z - length_2 );
						glVertex3f( center.x - length_2, center.y - length_2, center.z + length_2 );
						glVertex3f( center.x + length_2, center.y - length_2, center.z + length_2 );
						glVertex3f( center.x + length_2, center.y - length_2, center.z - length_2 );
						
						glVertex3f( center.x - length_2, center.y + length_2, center.z - length_2 );
						glVertex3f( center.x - length_2, center.y + length_2, center.z + length_2 );
						glVertex3f( center.x + length_2, center.y + length_2, center.z + length_2 );
						glVertex3f( center.x + length_2, center.y + length_2, center.z - length_2 );
						
						glVertex3f( center.x - length_2, center.y - length_2, center.z - length_2 );
						glVertex3f( center.x - length_2, center.y + length_2, center.z - length_2 );
						glVertex3f( center.x + length_2, center.y + length_2, center.z - length_2 );
						glVertex3f( center.x + length_2, center.y - length_2, center.z - length_2 );
						
						glVertex3f( center.x - length_2, center.y - length_2, center.z + length_2 );
						glVertex3f( center.x - length_2, center.y + length_2, center.z + length_2 );
						glVertex3f( center.x + length_2, center.y + length_2, center.z + length_2 );
						glVertex3f( center.x + length_2, center.y - length_2, center.z + length_2 );
					glEnd();
				}
				
				glEndList();
			}
			else
			{
				glCallList(displayList);	
			}
		}
		
		// draw victim location
		if( hasVictim > .9 )
		{
			float length_2 = edgeLength/3;
			
			glColor3f(.2,.2,1);
			
			glBegin(GL_LINES);
				glVertex3f( center.x, center.y, center.z + length_2 );
				glVertex3f( center.x, center.y, center.z - length_2 );
				
				glVertex3f( center.x, center.y + length_2, center.z );
				glVertex3f( center.x, center.y - length_2, center.z );
				
				glVertex3f( center.x + length_2, center.y, center.z );
				glVertex3f( center.x - length_2, center.y, center.z );
			glEnd();
			
			glColor3f(1,1,1);
		}
		
		// draw rfid location (s)?
	}
	else
	{
		for(int i=0; i<8; i++)
		if( children[i] != NULL )
			children[i]->drawGL();
	}
}
#endif
