#include "Util3D.h"

#ifdef USE_OPENGL
#include <QtOpenGl>
#endif

#include <QtDebug>

#include <limits>

QDataStream& operator<<(QDataStream& out, const Pose3D& pose)
{
	out << pose.x;
	out << pose.y;
	out << pose.z;
	
	out << pose.rotX;
	out << pose.rotY;
	out << pose.rotZ;
	
	return out;
}

QDataStream& operator>>(QDataStream& in, Pose3D& pose)
{
	in >> pose.x;
	in >> pose.y;
	in >> pose.z;
	
	in >> pose.rotX;
	in >> pose.rotY;
	in >> pose.rotZ;
	
	return in;
}

Pose3D operator-(const Pose3D& p1, const Pose3D& p2) {
	// the pose of p1 as seen from the position p2
	Matrix4 p2m;
	p2m.invFromPose3D(p2);
	
	Vector3 p1v(p1.x, p1.y, p1.z);
	Vector3 v = p2m.multiply(p1v);
	
	Pose3D ans;
	ans.x = v.x;
	ans.y = v.y;
	ans.z = v.z;
	
	ans.rotX = p1.rotX - p2.rotX;
	while(ans.rotX >= 2*M_PI)
		ans.rotX -= 2*M_PI;
	while(ans.rotX < 0)
		ans.rotX += 2*M_PI;
		
	ans.rotY = p1.rotY - p2.rotY;
	while(ans.rotY >= 2*M_PI)
		ans.rotY -= 2*M_PI;
	while(ans.rotY < 0)
		ans.rotY += 2*M_PI;
		
	ans.rotZ = p1.rotZ - p2.rotZ;
	while(ans.rotZ >= 2*M_PI)
		ans.rotZ -= 2*M_PI;
	while(ans.rotZ < 0)
		ans.rotZ += 2*M_PI;
		
	return ans;
}

Pose3D operator+(const Pose3D& p1, const Pose3D& p2) {
	// pose reached by p1 when seen from p2
	
	Matrix4 p2m;
	p2m.fromPose3D(p2);
	
	Vector3 p1v(p1.x, p1.y, p1.z);
	Vector3 v = p2m.multiply(p1v);
	
	Pose3D ans;
	ans.x = v.x;
	ans.y = v.y;
	ans.z = v.z;
	
	ans.rotX = p1.rotX + p2.rotX;
	while(ans.rotX >= 2*M_PI)
		ans.rotX -= 2*M_PI;
	while(ans.rotX < 0)
		ans.rotX += 2*M_PI;
		
	ans.rotY = p1.rotY - p2.rotY;
	while(ans.rotY >= 2*M_PI)
		ans.rotY -= 2*M_PI;
	while(ans.rotY < 0)
		ans.rotY += 2*M_PI;
		
	ans.rotZ = p1.rotZ - p2.rotZ;
	while(ans.rotZ >= 2*M_PI)
		ans.rotZ -= 2*M_PI;
	while(ans.rotZ < 0)
		ans.rotZ += 2*M_PI;
		
	return ans;
}

void Matrix4::loadIdentity() {
	for(int i=0; i<4; i++)
		for(int j=0; j<4; j++)
			if(i!=j)
				nums[index(i,j,4)] = 0;
			else
				nums[index(i,j,4)] = 1;
}

void Matrix4::rotate(float rx, float ry, float rz) {
	Matrix4 rot;
	rot.setRotation(rx,ry,rz);
	
	(*this) = rot.multiply((*this));
}

void Matrix4::translate(float x, float y, float z) {
	Matrix4 t;
	t.setTranslation(x,y,z);
	
	(*this) = t.multiply((*this));
}

void Matrix4::invRotate(float rx, float ry, float rz) {
	Matrix4 rot;
	rot.setInvRotation(rx,ry,rz);
	
	(*this) = this->multiply(rot);
}

void Matrix4::invTranslate(float x, float y, float z) {
	Matrix4 t;
	t.setTranslation(-x,-y,-z);
	
	(*this) = this->multiply(t);
}

void Matrix4::setRotation(float rx, float ry, float rz) {
	float cx = cos(rx);
	float sx = sin(rx);
	float cy = cos(ry);
	float sy = sin(ry);
	float cz = cos(rz);
	float sz = sin(rz);
	
	(*this)(0,0) = cy * cz;
	(*this)(0,1) = -sz * cy;
	(*this)(0,2) = sy;
	(*this)(1,0) = cz * sy * sx + sz * cx;
	(*this)(1,1) = cz * cx - sz * sy * sx;
	(*this)(1,2) = -cy * sx;
	(*this)(2,0) = sz * sx - cz * sy * cx;
	(*this)(2,1) = sz * sy * cx + sx * cz;
	(*this)(2,2) = cy * cx;
}

void Matrix4::setInvRotation(float rx, float ry, float rz) {
	float cx = cos(rx);
	float sx = sin(rx);
	float cy = cos(ry);
	float sy = sin(ry);
	float cz = cos(rz);
	float sz = sin(rz);
	
	// actually transpose of rotation matrix
	
	(*this)(0,0) = cy * cz;
	(*this)(0,1) = cz * sy * sx + sz * cx;
	(*this)(0,2) = -cz * sy * cx + sz * sx;
	(*this)(1,0) = -sz * cy;
	(*this)(1,1) = -sz * sy * sx + cz * cx;
	(*this)(1,2) = sz * sy * cx + cz * sx;
	(*this)(2,0) = sy;
	(*this)(2,1) = -cy * sx;
	(*this)(2,2) = cy * cx;
}

void Matrix4::setTranslation(float x, float y, float z) {
	(*this)(0,3) = x;
	(*this)(1,3) = y;
	(*this)(2,3) = z;
}

Vector3 Matrix4::multiply(const Vector3 &v) {
	Vector3 ans;
	ans.x = v.x*(*this)(0,0) + v.y*(*this)(0,1) + v.z*(*this)(0,2) + (*this)(0,3);
	ans.y = v.x*(*this)(1,0) + v.y*(*this)(1,1) + v.z*(*this)(1,2) + (*this)(1,3);
	ans.z = v.x*(*this)(2,0) + v.y*(*this)(2,1) + v.z*(*this)(2,2) + (*this)(2,3);
	
	return ans;
}

Matrix4 Matrix4::multiply(const Matrix4 &m) {
	Matrix4 ans;
	
	double sum=0;
	for(int i=0; i<4; i++) {
		for(int j=0; j<4; j++) {
			sum=0;
			for(int k=0; k<4; k++)
				sum += (*this)(i,k) * m(k,j);
			ans(i,j) = sum;
		}
	}
	
	return ans;
}

Matrix4 Matrix4::operator*(const Matrix4& m) {
	return this->multiply(m);
}
Vector3 Matrix4::operator*(const Vector3& m) {
	return this->multiply(m);
}

Matrix4 Matrix4::operator+(const Matrix4& m) {
	Matrix4 ans;
	for(int i=0; i<16; i++)
		ans.nums[i] = nums[i] + m.nums[i];
	return ans;
}
Matrix4 Matrix4::operator-(const Matrix4& m) {
	Matrix4 ans;
	for(int i=0; i<16; i++)
		ans.nums[i] = nums[i] - m.nums[i];
	return ans;
}

Matrix4 Matrix4::transpose() {
	Matrix4 ans;
	for(int i=0; i<4; i++)
		for(int j=0; j<4; j++)
			ans(i,j) = (*this)(j,i);
	return ans;
}

Matrix4 Matrix4::inverse() {
	bool is2by2 = true;
	for(int i=0; i<4; i++) {
		for(int j=0; j<4; j++) {
			if( ( i>=2 || j>=2 ) && ( i!=j ) && nums[ index(i,j,4) ] != 0 )
				is2by2 = false;
			if( ( i>=2 || j>=2 ) && ( i==j ) && nums[ index(i,j,4) ] != 1 )
				is2by2 = false;
		}
	}
	
	if(!is2by2)
		throw Exception("Can't invert a non-2x2 matrix!");
	
	Matrix4 ans;
	
	float det = (*this)(0,0) * (*this)(1,1) - (*this)(0,1) * (*this)(0,1);
    ans(0,0) = 1/det * (*this)(1,1);
    ans(0,1) = 1/det * -(*this)(0,1);
    ans(1,0) = 1/det * -(*this)(1,0);
    ans(1,1) = 1/det * (*this)(0,0);
    return ans;
}

Matrix4 & Matrix4::operator=(const Matrix4& m) {
	//qDebug() << "Matrix4 assgn!";
	for(int i=0; i<16; i++)
		nums[i] = m.nums[i];
	
	return *this;
}

Vector3 operator*(const Vector3 & v, const Matrix4 & m) {
	Vector3 ans;
	ans.x = v.x * m(0,0) + v.y * m(1,0) + v.z * m(2,0);
	ans.y = v.x * m(0,1) + v.y * m(1,1) + v.z * m(2,1);
	ans.z = v.x * m(0,2) + v.y * m(1,2) + v.z * m(2,2);
	return ans;
}
Matrix4 operator*(const float & f, const Matrix4 & m) {
	Matrix4 ans;
	for(int i=0; i<16; i++)
		ans.nums[i] = f * m.nums[i];
	return ans;
}

// *****************************************************************************************
// * FROM the book "Computer Graphics using Open GL", by F.S. Hill, Jr., Pearson Education *
// *****************************************************************************************

#ifdef USE_OPENGL
Camera::Camera()
{
	viewAngle=aspect=nearDist=farDist=0.0;
	set(Point3(), Point3(1,0,0), Vector3(0,0,1));
}

void Camera :: setModelViewMatrix(void)
{ // load modelview matrix with existing camera values

	//qDebug("new camera eye: x:"+QString::number(eye.x)+" y:"+QString::number(eye.y)+" z:"+QString::number(eye.z));
	float m[16];
	Vector3 eVec(eye.x, eye.y, eye.z); // a vector version of eye 
	m[0] =  u.x; m[4] =  u.y; m[8]  =  u.z;  m[12] = -eVec.dot(u);
	m[1] =  v.x; m[5] =  v.y; m[9]  =  v.z;  m[13] = -eVec.dot(v);
	m[2] =  n.x; m[6] =  n.y; m[10] =  n.z;  m[14] = -eVec.dot(n);
	m[3] =  0;   m[7] =  0;   m[11] =  0;    m[15] = 1.0;
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(m); // load OpenGLs modelview matrix
}

void Camera:: set(Point3 Eye, Point3 look, Vector3 up)
{
	// create a modelview matrix and send it to OpenGL
	eye.set(Eye); // store the given eye position
	n.set(eye.x - look.x, eye.y - look.y, eye.z - look.z); // make n
	u=(up.cross(n)); // make u = up X n
	n.normalize(); u.normalize(); // make them unit length
	v=(n.cross(u));  // make v =  n X u
	setModelViewMatrix(); // tell OpenGL
	
	// u = y
	// n = -x
	// v = z
}

void Camera :: roll(float angle)
{
	// u = y
	// n = -x
	// v = z
	
	// roll the camera through angle degrees
	float cs = cos(3.14159265/180 * angle);
	float sn = sin(3.14159265/180 * angle);
	
	Vector3 t(u); // remember old u
	u = cs*t + sn*v;
	v = -sn*t + cs*v;
	
	setModelViewMatrix();
}

void Camera :: pitch(float angle)
{
	// u = y
	// n = -x
	// v = z
	
	// roll the camera through angle degrees
	float cs = cos(3.14159265/180 * angle);
	float sn = sin(3.14159265/180 * angle);
	
	Vector3 t(n); // remember old u
	n = cs*t + sn*v;
	v = -sn*t + cs*v;
	
	setModelViewMatrix();
}

void Camera :: yaw(float angle)
{
	// u = y
	// n = -x
	// v = z
	
	// roll the camera through angle degrees
	float cs = cos(3.14159265/180 * angle);
	float sn = sin(3.14159265/180 * angle);
	
	Vector3 t(u); // remember old u
	u = cs*t + sn*n;
	n = -sn*t + cs*n;
	
	setModelViewMatrix();
}

void Camera::rotate(float angle)
{
	// u = y
	// n = -x
	// v = z
	
	// rotate only around z
	rotateZ(u,angle);
	rotateZ(n,angle);
	rotateZ(v,angle);
	
	setModelViewMatrix();
}

void Camera::rotateZ(Vector3& vec, float angle)
{
	float cs = cos(3.14159265/180 * angle);
	float sn = sin(3.14159265/180 * angle);
	
	float x = vec.x;
	float y = vec.y;
	
	vec.x = cs*x - sn*y;
	vec.y = cs*y + sn*x;
}


// slide( forward, right, up )
void Camera::slide(float delN, float delU, float delV) // slide it
{
	// u = y
	// n = -x
	// v = z
	
	eye.x += delU*u.x + delV*v.x - delN*n.x;
	eye.y += delU*u.y + delV*v.y - delN*n.y;
	eye.z += delU*u.z + delV*v.z - delN*n.z;
	
	setModelViewMatrix();
}

void Camera::setShape(float vAng, float asp, float nearD, float farD)
{
	viewAngle=vAng;
	aspect=asp;
	nearDist=nearD;
	farDist=farD;
	glMatrixMode(GL_PROJECTION);
	gluPerspective(vAng,asp,nearD,farD);
}

void Camera::setEye(float x, float y, float z)
{
	eye.set(x,y,z);
	setModelViewMatrix();
}

void Camera::move(float x, float y, float z)
{
	eye.x+=x;
	eye.y+=y;
	eye.z+=z;
	setModelViewMatrix();
}
#endif /* USE_OPENGL */


float Vector3::fmin()
{
	float min = x;
	if(min > y)
		min = y;
	if(min > z)
		return z;
	return min;
}

float Vector3::fmax()
{
	float max = x;
	if(max < y)
		max = y;
	if(max < z)
		return z;
	return max;
}

char Vector3::minIndex()
{
	float min = x;
	char mini = 0;
	if(min > y) {
		min = y;
		mini = 1;
	}
	if(min > z)
		return 2;
	return mini;
}

float Vector3::length() const
{
	return sqrt(x*x + y*y + z*z);
}

Vector3 operator*(const float & left, const Vector3 & right)
{
	Vector3 temp(left*right.x, left*right.y, left*right.z);
	return temp;
}

float operator*(const Vector3& left, const Vector3 & right)
{
    return left.dot(right);
}

Vector3 Vector3::operator-(const Vector3 & right)
{
	Vector3 temp(x-right.x,y-right.y,z-right.z);
	return temp;
}

Vector3 Vector3::operator+(const Vector3 & right)
{
	Vector3 temp(x+right.x,y+right.y,z+right.z);
	return temp;
}

float Vector3::divide(float nom, float denom)
{
	if( denom != 0 )
		return nom/denom;
	
	if( nom >= 0 )
		return std::numeric_limits<float>::infinity();
	else
		return -std::numeric_limits<float>::infinity();
}

Vector3 Vector3::operator/(const Vector3 & right)
{	
	Vector3 temp(divide(x,right.x),divide(y,right.y),divide(z,right.z));
	return temp;
}

Vector3 Vector3::operator/(const float & right)
{
	Vector3 temp(x/right,y/right,z/right);
	return temp;
}

Vector3& Vector3::operator=(const Vector3 & right)
{
	x=right.x;
	y=right.y;
	z=right.z;
	return *this;
}
