#ifndef SCANUTILS_H_
#define SCANUTILS_H_

#include <QtDebug>
#include <QUuid>
#include <time.h>
#include <math.h>

#include "Exception.h"
#include "LaserData.h"
#include "LaserData3D.h"
#include "RFIDData.h"
#include "VictimRFIDData.h"
#include "StatusData.h"
#include "Util3d.h"

class ScanObservation;
class RFIDObservation;
class VictimObservation;
class MatchResult;
class Matrix2;
class Matrix3;
class Point2;
class Vector2;

struct Pose2D {
	float x;
	float y;
	float rot;
};

struct Bounds  {
    float top;
    float left;
    float height;
    float width;
};
Bounds operator+(const Bounds&, const Bounds&);


class Random {
public:
    static void seed() {
        time_t secs = time(NULL);
        srand((int) secs);
    }

    static bool flipCoin() {
        return (generate(0, 1) == 1);
    }
    static int generate(int high) {
        return generate(0, high);
    }
    static int generate(int low, int high) {
        double r = (   (double)rand() / ((double)(RAND_MAX)+(double)(1)) );
        double x = (r * (double) (high-low));
        int    y = floor(x + (double) low + 0.5f);
        return y;
    }
    static double generate(double low, double high) {
    		double r = (   (double)rand() / ((double)(RAND_MAX)+(double)(1)) );
        double x = (r * (double) (high-low));
        double  y = floor(x + (double) low);
        return y;
    }
};

class Matrix2 {

public:

    float nums[4];


	Matrix2(float a, float b, float c, float d) {
        nums[0] = a;
        nums[1] = b;
        nums[2] = c;
        nums[3] = d;
    }

    Matrix2(float n) {
		for (int i = 0; i < 4; i++)
			nums[i] = n;
    }
    Matrix2(float m[4]) {
		for (int i = 0; i < 4; i++)
			nums[i] = m[i];
	}

    Matrix2() {
		loadIdentity();	
	}

    Matrix2(const Matrix2& m) {
		for (int i = 0; i < 4; i++)
			nums[i] = m.nums[i];
	}
	
    void loadIdentity();

    Matrix2 inverse();
    Matrix2 transpose();
	
	float& operator()(int i, int j) {
		//ith row, jth col
		if (i >= 2 || j >= 2)
			throw Exception("Out of bounds!");

		return nums[index(i, j, 2)];
	}
	
	const float& operator()(int i, int j) const {
		//ith row, jth col
		if (i >= 2 || j >= 2)
			throw Exception("Out of bounds!");

		return nums[index(i, j, 2)];
	}

	Matrix2 multiply(Matrix2 &);
	Vector2 multiply(Vector2 &);
	
	Matrix2 operator*(Matrix2& right);
	Vector2 operator*(const Vector2& right);
    Matrix2 operator-(const Matrix2 & right);
	Matrix2 operator+(const Matrix2 & right);
	Matrix2& operator=(const Matrix2&);
	
private:

	int index(int i, int j, int size) const {
		return i*size + j;
	}

};
Matrix2 operator*(const float & left, const Matrix2 & right);
Vector2 operator*(const Vector2 & left, const Matrix2 & right);


class Matrix3 {

public:

    float nums[9];

    Matrix3(float f) {
		for (int i = 0; i < 9; i++)
			nums[i] = f;
	}

    Matrix3(float m[9]) {
		for (int i = 0; i < 9; i++)
			nums[i] = m[i];
	}

    Matrix3() {
		loadIdentity();	
	}

    Matrix3(const Matrix3& m) {
		for (int i = 0; i < 9; i++)
			nums[i] = m.nums[i];
	}
	
	Matrix3 operator*(const Matrix3& right);
	Matrix3 operator+(const Matrix3 & right);

    void loadIdentity();
    Matrix3 transpose();
	
	// use these for consecutive actions
	void translate(float, float);
	void rotate(float);
	void invTranslate(float, float);
	void invRotate(float);
	
	// one shot setters
	void setRotation(float angle);
	void setInvRotation(float angle);
	void setTranslation(float x, float y);
	
	void fromPose2D(Pose2D& pose) {
		translate(pose.x, pose.y);
		rotate(pose.rot);
	}
	
	void invFromPose2D(Pose2D& pose) {
		invRotate(pose.rot);
		invTranslate(pose.x, pose.y);
	}
	
	float& operator()(int i, int j) {
		//ith row, jth col
		if (i >= 3 || j >= 3)
			throw Exception("Out of bounds!");

		return nums[index(i, j, 3)];
	}
	
	const float& operator()(int i, int j) const {
		//ith row, jth col
		if (i >= 3 || j >= 3)
			throw Exception("Out of bounds!");

		return nums[index(i, j, 3)];
	}

	Vector2 multiply(const Vector2 &);
	Matrix3 multiply(Matrix3 &);
	
	Matrix3& operator=(const Matrix3&);
	
private:

	int index(int i, int j, int size) const {
		return i*size + j;
	}

};



class Point2 { 

public: 
	float x, y;
	void set(float dx, float dy) { x = dx;  y = dy;  }
	void set(Point2& p)          { x = p.x; y = p.y; } 	
	Point2(float xx, float yy)   { x = xx;  y = yy;  }
	Point2()                     { x = 0;   y = 0;   }

};

class Vector2 { 

public: 
	float x, y;

    void set(float dx, float dy) { x = dx;  y = dy;  } 
	void set(const Vector2& v)   { x = v.x; y = v.y; }
	void flip()                  { x = -x;  y = -y;  } // reverse this vector
	void setDiff(Point2& a, Point2& b) //set to difference a - b
	    { x = a.x - b.x; y = a.y - b.y; }
	void normalize() //adjust this vector to unit length
    {
		double sizeSq = x*x + y*y;
		if (sizeSq < 0.0000001)
		{
			return; // does nothing to zero vectors;
		}
		float scaleFactor = 1.0 / (float)sqrt(sizeSq);
		x *= scaleFactor; 
        y *= scaleFactor; 
	}
	Vector2(float xx, float yy) {x = xx;  y = yy;  }
	Vector2(const Vector2& v)   {x = v.x; y = v.y; }
	Vector2()                   {x = 0;   y = 0;   }

    Vector2  operator-(const Vector2 & right);
	Vector2  operator+(const Vector2 & right);
	Vector2& operator=(const Vector2 & right);
};
Vector2 operator*(const float & left, const Vector2 & right);
float operator*(const Vector2 & left, const Vector2 & right);

class ScanObservation
{
    
public:

	// for replication
	ScanObservation(QUuid);
	
    ScanObservation(float unit, float x, float y, float rot, LaserData laser);
    ScanObservation(float unit, Pose2D rawpose, Pose2D pose, float* range, double minDistance, double maxDistance, uint length);
    virtual ~ScanObservation();

    QUuid  getUniqueID();

    void addRFID(RFIDData rfid);
    void addVictim(VictimRFIDData victim);
    
    int countRFIDs() {
    		return rfids.size();
    }
    
    int countVictims() {
    		return victims.size();
    }
    
	RFIDData getRFID(QString ID) {
		return rfids[ID];
	}
    bool hasRFIDs() {
    		return rfids.size() != 0;	
    }
    
	VictimRFIDData getVictim(QString victimID) {
		return victims[victimID];
	}
    bool hasVictims() {
    		return victims.size() != 0;	
    }

    QMapIterator< QString, RFIDData >   rfidIterator();
    QMapIterator< QString, VictimRFIDData > victimIterator();

    Pose2D rawpose;
	Pose2D pose;
    float* range;
    float  unit;
    uint   length;
	double mindistance;
	double maxdistance;

private:
    QUuid  uniqueID;

    QMap< QString, RFIDData >   rfids;
    QMap< QString, VictimRFIDData > victims;

};


class MatchResult
{
    
public:

    MatchResult(Pose2D rawdpose, Pose2D smdpose, Matrix3 mcov, uint numIterations, uint numCorrespondencies, uint numPatchesInSubmap);

	Pose2D  rawdpose; // original pose estimate (from odometry / log file)
	Pose2D  smdpose;  // the scanmatcher's odometry estimate
    Matrix3 mcov;     // matching covariance

    uint    numIterations;
    uint    numCorrespondencies;
    uint    numPatchesInSubmap;

};

#endif 
