#include "WeightedScanMatcher.h"

#include <QDebug>


// only results below the error threshold are considered
const float WeightedScanMatcher::ERROR_THRESHOLD = .01F;


// incidence angle filter
const float WeightedScanMatcher::IANGLE_THRESHOLD  = M_PI/8;

// for tuning the noise error covariance
const float WeightedScanMatcher::DEFAULT_SIGMAANG  = .002F;
const float WeightedScanMatcher::DEFAULT_SIGMADIST = 10.0F;  

// for tuning the corresponde error covariance
const float WeightedScanMatcher::ANGLE_INCREMENT   = M_PI/360;
const float WeightedScanMatcher::SIGMA_CORRELATED  = 10.0F;


WeightedScanMatcher::WeightedScanMatcher()
: ScanMatcher()
{
}

WeightedScanMatcher::~WeightedScanMatcher()
{
}

MatchResult* WeightedScanMatcher::matchScans(QList<ScanObservation*>* submap, ScanObservation* primary, ScanObservation* scan2) {

	this->NUMSCANS = (submap->count() > 4) ? 4 : submap->count();
    this->LENGTH   = scan2->length;
    this->LENGTH1  = this->LENGTH * this->NUMSCANS;
    this->LENGTH2  = this->LENGTH;

	this->RANGESCAN_MINDISTANCE = scan2->mindistance;
	this->RANGESCAN_MAXDISTANCE = scan2->maxdistance;
	ScanObservation* scan1;
    for (uint s = 0; s < (uint) submap->count(); s++) {
        scan1 = submap->at(s);
		if (scan1->mindistance > this->RANGESCAN_MINDISTANCE) {
			this->RANGESCAN_MINDISTANCE = scan1->mindistance;
		}
		if (scan1->maxdistance < this->RANGESCAN_MAXDISTANCE) {
			this->RANGESCAN_MAXDISTANCE = scan1->maxdistance;
		}
	}

    float**  angles  = this->computeAngles(-M_PI_2, M_PI_2, M_PI/(this->LENGTH - 1.0));
    Vector2* points1 = this->toPoints(submap, primary, angles);
    Vector2* points2 = this->toPoints(scan2, angles);
    Pose2D   dpose   = this->getRelativePose(primary->rawpose, scan2->rawpose);

    MatchResult* result = this->matchPoints(points1, points2, dpose);

    // CLEANUP
    for (uint i = 0; i < LENGTH; i++) {
        delete[] angles[i];
    }
    delete[] angles;
    delete[] points1;
    delete[] points2;
    angles = 0;
    points1 = points2 = 0;

    return result;

}

MatchResult* WeightedScanMatcher::matchScans(ScanObservation* scan1, ScanObservation* scan2)
{
    if (scan1->length != scan2->length)
        throw Exception("Error: scans must be of same length");
    if (scan1->length < 2 || scan2->length < 2)
        throw Exception("Error: cannot match scans with fewer than 2 scanpoints");

	this->RANGESCAN_MINDISTANCE = 
		(scan1->mindistance > scan2->mindistance) ?
		scan1->mindistance : scan2->mindistance;
	this->RANGESCAN_MAXDISTANCE = 
		(scan1->maxdistance < scan2->maxdistance) ?
		scan1->maxdistance : scan2->maxdistance;

    this->NUMSCANS = 1;
    this->LENGTH   = scan1->length;
    this->LENGTH1  = this->LENGTH;
    this->LENGTH2  = this->LENGTH;

    float**  angles  = this->computeAngles(-M_PI_2, M_PI_2, M_PI/(this->LENGTH - 1));
    Vector2* points1 = this->toPoints(scan1, angles);
    Vector2* points2 = this->toPoints(scan2, angles);
    Pose2D   dpose   = this->getRelativePose(scan1->rawpose, scan2->rawpose);

    MatchResult* result = this->matchPoints(points1, points2, dpose);

    // CLEANUP
    for (uint i = 0; i < LENGTH; i++) {
        delete[] angles[i];
    }
    delete[] angles;
    delete[] points1;
    delete[] points2;
    angles = 0;
    points1 = points2 = 0;

    return result;
}

MatchResult* WeightedScanMatcher::matchPoints(Vector2* points1, Vector2* points2, Pose2D odometry) {

    bool*   filter1  = new bool[LENGTH1];
    bool*   filter2  = new bool[LENGTH2];
    float   dist;
    for (uint i = 0; i < LENGTH1; i++) {
        dist = this->toDistance(points1[i]);
        filter1[i] = 
            dist < this->RANGESCAN_MINDISTANCE ||
            dist > this->RANGESCAN_MAXDISTANCE;
    }
    for (uint i = 0; i < LENGTH2; i++) {
        dist = this->toDistance(points2[i]);
        filter2[i] = 
            dist < this->RANGESCAN_MINDISTANCE ||
            dist > this->RANGESCAN_MAXDISTANCE;
    }

    float** iangles1 = this->computeIncidenceAngles(points1, this->NUMSCANS);
    float** iangles2 = this->computeIncidenceAngles(points2, 1);

    // J ...
    Matrix2  J(0.0F, -1.0F, 1.0F, 0.0F);
    Matrix2  Jt = J.transpose();

    // the key variables 
    Pose2D   rawdpose, dpose;
    Vector2  midpoint, transest;
    double   rotest;
    Matrix3  rotmx;
    float    results[this->RESULT_COUNT];
    uint     errorcount;



    // temporary usage vars
    Vector2* newpoints2;
    uint**   corr;
    uint     corrcount, p1, p2;
    Vector2  point1, point2, dpoint, vdpose;
    Matrix2  cov1, cov2, sumcov, invsumcov, sumsumcov, irotmid;
    double   nval, dval;
    bool     mark;
    
    // summarized vars    
    Matrix2  suminvsumcov, sumirotmid;
    Vector2  sumipoint1, sumipoint2, sumrotmid;
    double   sumerror, sumnval, sumdval;


    // init ...
    rawdpose   = odometry;
    dpose      = odometry;
    errorcount = 0;
    for (uint i = 0; i < this->RESULT_COUNT; i++)
        results[i] = this->ERROR_THRESHOLD + 1.0F;

    // main loop
    uint n;
    for (n = 1; n <= MAXITERATIONS; n++) {
        // transform the points of scan2 to the frame of scan1 using dpose
        newpoints2 = this->toPose(points2, dpose);
        corr       = this->correlatePoints(points1, newpoints2, filter1, filter2, corrcount);

        if (corrcount > 0) {

            // reset 
            suminvsumcov = Matrix2(0.0F);
            sumipoint1   = Vector2(0.0F, 0.0F);
            sumipoint2   = Vector2(0.0F, 0.0F);
            sumerror     = 0.0;
            sumirotmid   = Matrix2(0.0F);
            sumrotmid    = Vector2(0.0F, 0.0F);

            for (uint c = 0; c < corrcount; c++) {

                // get corresponding points and error covariances ...
                p1        = corr[c][0];
                p2        = corr[c][1];
                point1    = points1[p1];
                point2    = newpoints2[p2];
                cov1      = this->computeErrorCovariance(point1, iangles1[p1]);
                cov2      = this->computeErrorCovariance(point2, iangles2[p2]);

                // translation estimate ...
                sumcov    = cov1 + cov2;
                invsumcov = sumcov.inverse();

                suminvsumcov = suminvsumcov + invsumcov;
                sumipoint1   = sumipoint1   + invsumcov * point1;
                sumipoint2   = sumipoint2   + invsumcov * point2;


                // error metric ...
                dpoint       = point1 - point2;
                sumerror     = sumerror + (double) (dpoint * invsumcov * dpoint);

                // rotation estimate ...
                irotmid      = Jt * invsumcov * J;
                sumirotmid   = sumirotmid + 2 * irotmid;
                sumrotmid    = sumrotmid + irotmid * point1 + irotmid * point2;

            }

            // get rotational midpoint and translation estimate
            midpoint  = sumirotmid.inverse() * sumrotmid;
            sumsumcov = suminvsumcov.inverse(); 
            transest  = sumsumcov * (sumipoint2 - sumipoint1);



            // reset 
            sumnval   = 0.0;
            sumdval   = 0.0;

            for (uint c = 0; c < corrcount; c++) {

                // get corresponding points and error covariances ...
                p1        = corr[c][0];
                p2        = corr[c][1];
                point1    = points1[p1];
                point2    = newpoints2[p2];
                cov1      = this->computeErrorCovariance(point1, iangles1[p1]);
                cov2      = this->computeErrorCovariance(point2, iangles2[p2]);

                // apply estimated midpoint and translation
                point1    = point1 - midpoint;
                point2    = point2 - transest - midpoint;

                sumcov    = cov1 + cov2;
                invsumcov = sumcov.inverse();

                dpoint    = point2 - point1;

                // update numerator and denominator summaries
                nval      = dpoint * invsumcov * J * point2;
                dval      = dpoint * invsumcov * point2 + point2 * J * invsumcov * J * point2;
                sumnval   = sumnval + nval;
                sumdval   = sumdval + dval;

            }

            // get rotational estimate
            rotest = -sumnval / sumdval;

            
            // update odometry estimate
            rotmx.setRotation(-rotest);
            vdpose.x  = dpose.x;
            vdpose.y  = dpose.y;
            vdpose    = vdpose - transest - midpoint;
            vdpose    = rotmx.multiply(vdpose);
            vdpose    = vdpose + midpoint;
            dpose.x   = vdpose.x;
            dpose.y   = vdpose.y;
            dpose.rot = dpose.rot - rotest;

			//qDebug() << QString("%1: dpose = %2,%3,%4").arg(n).arg(dpose.x).arg(dpose.y).arg(dpose.rot);


            // check/update break conditions
            mark = false;
            for (uint c = 0; c < this->RESULT_COUNT - 1; c++) {
                results[c] = results[c+1];
                if (sumerror == 0)
                    mark  = mark || (results[c] < this->ERROR_THRESHOLD);
                else
                    mark  = mark || (abs(results[c] / sumerror - 1) < this->ERROR_THRESHOLD);
            }
            results[this->RESULT_COUNT-1] = sumerror;

            if (mark)
                errorcount++;

        }

        if (errorcount >= this->RESULT_BREAKCOUNT || corrcount == 0 || n == this->MAXITERATIONS) {
            // don't clean up the results from this iteration
            // we need it later
            break;
        } else {
            // CLEANUP
            for (uint c = 0; c < corrcount; c++)
                delete[] corr[c];
            delete[] corr;
            corr = 0;
            delete[] newpoints2;
            newpoints2 = 0;
        }
    }





    double rtval, sumrtval = 0;
    Vector2 jqval, sumjqval(0.0F, 0.0F);
    suminvsumcov = Matrix2(0, 0, 0, 0);

    // compute match covariance
    for (uint c = 0; c < corrcount; c++) {

        // get corresponding points and error covariances ...
        p1        = corr[c][0];
        p2        = corr[c][1];
        point1    = points1[p1];
        point2    = newpoints2[p2];
        cov1      = this->computeErrorCovariance(point1, iangles1[p1]);
        cov2      = this->computeErrorCovariance(point2, iangles2[p2]);

        // apply estimated midpoint and translation
        point1       = point1 - midpoint;
        point2       = point2 - transest - midpoint;

        sumcov       = cov1 + cov2;
        invsumcov    = sumcov.inverse();
        suminvsumcov = suminvsumcov + invsumcov;

        // update rtval and jqval summaries
        rtval     = point2 * J * invsumcov * J * point2;
        sumrtval  = sumrtval + rtval;
        jqval     = invsumcov * J * point2;
        sumjqval  = sumjqval + jqval;

    }
    sumrtval = - (1 / sumrtval);
    
    Vector2 Prt = -sumrtval * suminvsumcov.inverse() * sumjqval;
    
    // the matching covariance
    Matrix3 mcov;
    mcov(0, 0) = sumsumcov(0, 0);
    mcov(0, 1) = sumsumcov(0, 1);
    mcov(1, 0) = sumsumcov(1, 0);
    mcov(1, 1) = sumsumcov(1, 1);
    mcov(0, 2) = mcov(2, 0) = Prt.x;
    mcov(2, 1) = mcov(1, 2) = Prt.y;
    mcov(2, 2) = sumrtval;

    Matrix3 S; // loads identity matrix by default
    S(0, 2) = midpoint.y;
    S(1, 2) = -midpoint.x;

    mcov = S * mcov * S.transpose();



    // CLEANUP
    for (uint i = 0; i < corrcount; i++)
        delete[] corr[i];
    delete[] corr;
    corr = 0;
    delete[] newpoints2;
    newpoints2 = 0;

    delete[] filter1;
    delete[] filter2;
    for (uint i = 0; i < LENGTH; i++) {
        delete[] iangles1[i];
        delete[] iangles2[i];
    }
    delete[] iangles1;
    delete[] iangles2;

    filter1 = filter2 = 0;
    iangles1 = iangles2 = 0;

    return new MatchResult(rawdpose, dpose, mcov, n, corrcount, this->NUMSCANS);

}


Pose2D WeightedScanMatcher::getRelativePose(Pose2D pose1, Pose2D pose2) {

    Pose2D dpose;

	//* USE ODOMETRY
    Matrix3 rotmx;
    Vector2 vdpose(pose2.x - pose1.x, pose2.y - pose1.y);
    rotmx.setRotation(-pose1.rot);
    vdpose = rotmx.multiply(vdpose);

    dpose.x   = vdpose.x;
    dpose.y   = vdpose.y;
    dpose.rot = pose2.rot - pose1.rot;
	//*/

	/* IGNORE ODOMETRY
    dpose.x   = 0.0F;
    dpose.y   = 0.0F;
    dpose.rot = 0.0F;
	//*/

	return dpose;
}

Vector2* WeightedScanMatcher::toPose(Vector2* points, Pose2D pose) {
    Matrix3 rotmx;
    rotmx.setRotation(pose.rot);
    rotmx.setTranslation(pose.x, pose.y);

    Vector2* newpoints = new Vector2[LENGTH];
    Vector2 point, newpoint;
    for (uint i = 0; i < LENGTH; i++) {
        point        = points[i];
        newpoint     = rotmx.multiply(point);
        newpoints[i] = newpoint;
    }
    return newpoints;
}

Vector2* WeightedScanMatcher::toPoints(QList<ScanObservation*>* scans, ScanObservation* primary, float** angles) {

    float distance;
    float angle;

    Vector2* points = new Vector2[LENGTH * scans->count()];
    ScanObservation* scan;
    Pose2D  dpose;

    for (uint s = 0; s < (uint) scans->count(); s++) {
        scan  = scans->at(s);
        dpose = this->getRelativePose(primary->pose, scan->pose);

        for (uint i = 0; i < LENGTH; i++) {
            distance = scan->range[i];
            angle    = angles[i][ANG] + dpose.rot;
            points[LENGTH*s + i] = Vector2(dpose.x + cos(angle) * distance, dpose.y + sin(angle) * distance);
        }
    }

    return points;
}

Vector2* WeightedScanMatcher::toPoints(ScanObservation* scan, float** angles) {
    
    float  distance;
    float* angle;

    Vector2* points = new Vector2[LENGTH];
    for (uint i = 0; i < LENGTH; i++) {
        distance  = scan->range[i];
        angle     = angles[i];
        points[i] = Vector2(angle[COS] * distance, angle[SIN] * distance);
    }

    return points;
}


float WeightedScanMatcher::toAngle(Vector2 point) {
    return atan2(point.y, point.x);
}

float WeightedScanMatcher::toDistance(Vector2 point) {
    return sqrt(pow(point.x, 2.0F) + pow(point.y, 2.0F));
}

uint** WeightedScanMatcher::correlatePoints(Vector2* points1, Vector2* points2, bool* filter1, bool* filter2, uint& corrcount) {

    float  maxdist = pow((float) this->CORRELATION_MAXDISTANCE, 2);

    Vector2 point1, point2;
    float dist;

    uint* potentials1  = new uint[LENGTH1];
    uint* potentials2  = new uint[LENGTH2];

    float** dists = new float*[LENGTH1];
    for (uint i = 0; i < LENGTH1; i++) {
        point1   = points1[i];
        dists[i] = new float[LENGTH2];
        for (uint j = 0; j < LENGTH2; j++) {
            point2      = points2[j];
            dist        = pow(point2.x - point1.x, 2) + pow(point2.y - point1.y, 2);
            dists[i][j] = dist;
        }
    }

    uint* pmarks1 = new uint[LENGTH1];
    uint  n2;
    for (uint i = 0; i < LENGTH1; i++) {
        dist = maxdist + 1;
        n2 = -1;
        if (!filter1[i]) {
            for (uint j = 0; j < LENGTH2; j++) {
                if (filter2[j])
                    continue;
                if (dists[i][j] < dist) {
                    dist = dists[i][j];
                    n2 = j;
                }
            }
        }
        if (dist < maxdist) {
            pmarks1[i] = n2;
        } else {
            pmarks1[i] = LENGTH2 + 1;
        }
    }


    uint  pairs = 0;

    uint* pmarks2 = new uint[LENGTH2];
    uint  n1;
    for (uint j = 0; j < LENGTH2; j++) {
        dist = maxdist + 1;
        n1 = -1;
        for (uint i = 0; i < LENGTH1; i++) {
            if (pmarks1[i] == j) {
                if (dists[i][j] < dist) {
                    dist = dists[i][j];
                    n1 = i;
                }
            }
        }
        if (dist < maxdist) {
            pmarks2[j] = n1;
            pairs++;
        } else {
            pmarks2[j] = LENGTH1 + 1;
        }
    }

    uint** corr = new uint*[pairs];
    uint c = 0;
    for (uint j = 0; j < LENGTH2; j++) {

        if (pmarks2[j] > LENGTH1)
            continue;

        n1 = pmarks2[j];
        n2 = j;

        corr[c] = new uint[2];
        corr[c][0] = n1;
        corr[c][1] = n2;
        c++;

        //qDebug() << "-- Correlated " << n1 << " <-> " << n2;

    }


    // CLEANUP
    for (uint i = 0; i < LENGTH1; i++)
        delete[] dists[i];
    delete[] dists;
    delete[] potentials1;
    delete[] potentials2;
    delete[] pmarks1;
    delete[] pmarks2;
    
    dists = 0;
    potentials1 = potentials2 = 0;
    pmarks1 = pmarks2 = 0;


    corrcount = pairs;
    return corr;
}


float** WeightedScanMatcher::computeAngles(float fromAngle, float toAngle, float resolution) {
    float fov = toAngle - fromAngle;
    int   len = fov / resolution + 1;

    if (len != LENGTH)
        throw Exception("Error: the specified field of view and resolution do not agree with the scan length");

    float** angles = new float*[LENGTH];
    float alpha;
    for (uint i = 0; i < LENGTH; i++) {
        alpha = fromAngle + resolution*i;
        angles[i] = new float[3];
        angles[i][ANG] = alpha;
        angles[i][COS] = cos(alpha);
        angles[i][SIN] = sin(alpha);
    }
    return angles;
}

float** WeightedScanMatcher::computeIncidenceAngles(Vector2* points, uint numScans) {

    Vector2* dpoints   = new Vector2[numScans * (LENGTH-1)];
    float*   dangles   = new float[numScans * (LENGTH-1)];
    Vector2  dpoint;
    for (uint s = 0; s < numScans; s++) {
        for (uint i = 0; i < LENGTH - 1; i++) {
            dpoint                    = points[LENGTH*s + i+1] - points[LENGTH*s + i];
            dpoints[(LENGTH-1)*s + i] = dpoint;
            dangles[(LENGTH-1)*s + i] = atan2(dpoint.y, dpoint.x);
        }
    }
    
    float*   cmpangles = new float[numScans * (LENGTH-2)];
    float*   avgangles = new float[numScans * (LENGTH-2)];
    float    cmpangle;
    int f;
    for (uint s = 0; s < numScans; s++) {
        for (uint i = 0; i < LENGTH - 2; i++) {
            cmpangle = dangles[(LENGTH-2)*s + i] - dangles[(LENGTH-2)*s + i+1];
            f = floor( (cmpangle+M_PI) / (2*M_PI) );
            cmpangle = cmpangle - 2 * M_PI * f;
            cmpangles[(LENGTH-2)*s + i] = cmpangle;
            avgangles[(LENGTH-2)*s + i] = dangles[(LENGTH-2)*s + i] - .5F * cmpangle;
        }
    }

    float**  iangles   = new float*[numScans * LENGTH];
    float iangle;

    for (uint s = 0; s < numScans; s++) {

        iangles[LENGTH*s] = new float[3];
        iangles[LENGTH*s][ANG] = 0;
        iangles[LENGTH*s][COS] = cos(0.0);
        iangles[LENGTH*s][SIN] = sin(0.0);
        iangles[LENGTH*s + LENGTH-1] = new float[3];
        iangles[LENGTH*s + LENGTH-1][ANG] = 0;
        iangles[LENGTH*s + LENGTH-1][COS] = cos(0.0);
        iangles[LENGTH*s + LENGTH-1][SIN] = sin(0.0);

		Vector2 point;
		float angle;
        for (uint i = 1; i < LENGTH - 1; i++) {
            if (abs(cmpangles[(LENGTH-2)*s + i-1]) < IANGLE_THRESHOLD) {
				point  = points[i];
				angle  = atan2(point.y, point.x);
                iangle = angle - avgangles[(LENGTH-2)*s + i-1];
                iangle = iangle - 2*M_PI * floor( (iangle+M_PI)   / (2*M_PI) );
                iangle = iangle -   M_PI * floor( (iangle+M_PI/2) / (M_PI)   );
            } else {
                iangle = 0;
            }

            iangles[LENGTH*s + i] = new float[3];
            iangles[LENGTH*s + i][ANG] = iangle;
            iangles[LENGTH*s + i][COS] = cos(iangle);
            iangles[LENGTH*s + i][SIN] = sin(iangle);
        }
    }

    // CLEANUP
    delete[] dpoints;
    delete[] dangles;
    delete[] cmpangles;
    delete[] avgangles;
    dpoints = 0;
    dangles = cmpangles = avgangles = 0;

    return iangles;
}

Matrix2 WeightedScanMatcher::computeErrorCovariance(Vector2 point, float* iangle) {
    Matrix2 ncov = this->computeNoiseErrorCovariance(point);
    Matrix2 ccov = this->computeCorrespondenceErrorCovariance(point, iangle);
    Matrix2 cov = ncov + ccov;
    return cov;
}

Matrix2 WeightedScanMatcher::computeNoiseErrorCovariance(Vector2 point) {

	float angle = this->toAngle(point);
	float dist  = this->toDistance(point);

    float c = cos(angle);
    float s = sin(angle);
    float a = pow(DEFAULT_SIGMADIST, 2);
    float b = pow(DEFAULT_SIGMAANG, 2) * pow(dist, 2);

    Matrix2 cov;
    cov(0,0) = pow(c,2)*a + pow(s,2)*b;
    cov(0,1) = c*s*a - c*s*b;
    cov(1,0) = cov(0,1);
    cov(1,1) = pow(s,2)*a + pow(c,2)*b;
    return cov;
}


Matrix2 WeightedScanMatcher::computeCorrespondenceErrorCovariance(Vector2 point, float* iangle) {

    float iang = iangle[ANG];
    if (iang == 0) {
        iang = this->IANGLE_THRESHOLD;
    }
	float angle = this->toAngle(point);
	float dist  = this->toDistance(point);
    float dmin  = (dist * sin(ANGLE_INCREMENT)) / sin(iang + ANGLE_INCREMENT);
    float dmax  = (dist * sin(ANGLE_INCREMENT)) / sin(iang - ANGLE_INCREMENT);
    float dsum  = dmin + dmax;

    float cerr = SIGMA_CORRELATED * (pow(dmin, 3) + pow(dmax, 3)) / (3 * dsum);

    Matrix2 cov;
    if (iangle[ANG] == 0) {
        cov(0,0) = cerr;
        cov(0,1) = 0;
        cov(1,0) = 0;
        cov(1,1) = cerr;

    } else {
        float dang = angle - iangle[ANG];
        float c    = cos(dang);
        float s    = sin(dang);

        cov(0,0) = pow(c,2) * cerr;
        cov(0,1) = c * s * cerr;
        cov(1,0) = cov(0,1);
        cov(1,1) = pow(s,2) * cerr;
    }

    return cov;
}


